/**************************************\
* nt2dump.c                            *
* Dump a NerdTracker II temp.dat file  *
*                                      ********************************\
* Copyright (C) 2000 Damian Yerrick                                    *
*                                                                      *
* This program is free software; you can redistribute it and/or modify *
* it under the terms of the GNU General Public License as published by *
* the Free Software Foundation; either version 2 of the License, or    *
* (at your option) any later version.                                  *
*                                                                      *
* This program is distributed in the hope that it will be useful,      *
* but WITHOUT ANY WARRANTY; without even the implied warranty of       *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
* GNU General Public License for more details.                         *
*                                                                      *
* You should have received a copy of the GNU General Public License    *
* along with this program; if not, write to the Free Software          *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*                                                                      *
\**********************************************************************/

#include <stdio.h>
#include <stdlib.h>

typedef unsigned char ui8;
typedef unsigned short ui16;
typedef unsigned long ui32;

char noteNames[]  = "CCDDEFFGGAAB";
char noteSharps[] = "-#-#--#-#-#-";

typedef struct NEDSeqHeader
{
  ui8 version; /* 0x20 */
  ui8 headerSize; /* 0x10 */
  ui8 flags; /* what the heck is this? */
  ui8 initSpeed; /* 0x06 */
  ui8 loopPoint;
  ui8 nInsts; /* <= 0x10 */
  ui8 nSamples;
  ui8 songLength;
  ui8 nPatterns[5];
} NEDSeqHeader;

typedef struct NEDInstrument
{
  ui8 reg0args;
  ui8 pitchBend;
  ui8 fm;
  ui8 am;
  ui8 noteLen;
  ui8 decay;
  ui8 arpX;
  ui8 arpYZ;
} NEDInstrument;

NEDSeqHeader hdr;
NEDInstrument insts[16];
ui32 *orders;

int patptrs[5][32];

ui8 patdata[8192];



void PrintHeader(void)
{
  printf("version:        $%02x\n", (int)hdr.version);
  printf("header size:    %3d\n", (int)hdr.headerSize);
  printf("flags:          $%02x\n", (int)hdr.flags);
  printf("ticks per row:  %3d\n", (int)hdr.initSpeed);
  printf("loop point:     %3d\n", (int)hdr.loopPoint);
  printf("number of insts:%3d\n", (int)hdr.nInsts);
  printf("num. of dmcinst:%3d\n", (int)hdr.nSamples);
  printf("song length:    %3d\n", (int)hdr.songLength);
  printf("square 1 pats:  %3d\n", (int)hdr.nPatterns[0]);
  printf("square 2 pats:  %3d\n", (int)hdr.nPatterns[1]);
  printf("triangle pats:  %3d\n", (int)hdr.nPatterns[2]);
  printf("noise patterns: %3d\n", (int)hdr.nPatterns[3]);
  printf("dmc patterns:   %3d\n", (int)hdr.nPatterns[4]);
}

int ReadOrders(ui32 *orders, int songLength, FILE *fp)
{
  int nRead = 0;
  ui8 orderStr[4];

  while(songLength > 0)
  {
    if(fread(orderStr, 4, 1, fp) < 1)
      return nRead;
    orders[nRead] = orderStr[0] + orderStr[1]*0x00000100L
                                + orderStr[2]*0x00010000L
                                + orderStr[3]*0x01000000L;
    nRead++;
    songLength--;
  }
  return nRead;
}

int ReadPatPtrs(FILE *fp)
{
  int i, j;
  
  ui8 orderStr[2];

  for(i = 0; i < 5; i++)
  {
    for(j = 0; j < hdr.nPatterns[i]; j++)
    {
      if(fread(orderStr, 2, 1, fp) < 1)
        return -1;
      patptrs[i][j] = orderStr[0] + orderStr[1]*0x00000100L;
      printf("patptrs[%d][%d] = $%04x\n", i, j, patptrs[i][j]);
    }
  }
  return 0;
}

void PrintInsts()
{
  int i;

  for(i = 0; i < hdr.nInsts; i++)
  {
    printf("\nInstrument %d:\n", i + 1);
    printf("waveform:       %3d\n", insts[i].reg0args >> 6);
    printf("hw decay loop:    %c\n", ((insts[i].reg0args >> 5) & 1) ? 'T' : 'F');
    printf("no hw decay:      %c\n", ((insts[i].reg0args >> 4) & 1) ? 'T' : 'F');
    printf("reg 0 d3-d0:    $%02x\n", insts[i].reg0args & 0x03);
    printf("initial volume: %3d\n", (insts[i].decay >> 2) & 0x3c);
    printf("note length:    $%02x\n", insts[i].noteLen & 0x7f);
    printf("arpeggio:      %c%x%x%x\n",
           (insts[i].arpX & 0x80) ? '-' : '+',
           insts[i].arpYZ & 0x0f,
           insts[i].arpYZ >> 4,
           insts[i].arpX & 0x0f);
    printf("fm speed:       %3d\n", insts[i].fm >> 4);
    printf("fm depth:       %3d\n", insts[i].fm& 0x0f);
    printf("am speed:       %3d\n", insts[i].am >> 4);
    printf("am depth:       %3d\n", insts[i].am & 0x0f);
    printf("sw decay:       %c%2d\n",
           ((insts[i].noteLen >> 7) & 1) ? '+' : '-',
           insts[i].decay & 0x0f);
    printf("alt. noise wvfm:  %c\n", ((insts[i].arpX >> 6) & 1) ? 'T' : 'F');
    printf("sw pitch bend:  %+3d\n", -(signed char)(insts[i].pitchBend));
    printf("arpX d4:        $%02x\n", ((insts[i].arpX >> 4) & 1));
    printf("nonlooped arp:    %c\n", ((insts[i].arpX >> 5) & 1) ? 'T' : 'F');
  }
}

void PrintOrders()
{
  int i, j;

  puts("\nOrder table:\n"
       "#   s1 s2 tr no dm");

  for(i = 0; i < hdr.songLength; i++)
  {
    int curO = orders[i];

    printf("%02x:", i);
    for(j = 0; j < 5; j++)
    {
      printf("|%02x", curO & 0x1f);
      curO >>= 5;
    }
    putchar('\n');
  }
}

void HexDump()
{
  int i, j;

  for(i = 0; i < 8192; i += 16)
  {
    printf("%4x:", i);
    for(j = i; j < i + 16; j++)
      printf(" %02x", patdata[j]);
    putchar('\n');
  }
}

void PrintTable(int order)
{
  int curpatptr[5] = {0};
  int patbitsptr[5] = {0};
  int rowsLeft[5] = {0};
  int nibble[5] = {0}; /* 0 or 4 */
  int row, col, gotbyte;

  int whichBits;

  for(col = 0; col < 5; col++)
  {
    whichBits = (orders[order] >> (5 * col)) & 0x1f;
    curpatptr[col] = patptrs[col][whichBits];
    if(curpatptr[col] != 0) /* do we have an Empty Pattern? */
      rowsLeft[col] = (signed char)(patdata[curpatptr[col] + 1]);
    else
      rowsLeft[col] = -1;
    curpatptr[col] += 2;
    patbitsptr[col] = curpatptr[col];
    curpatptr[col] += 1 + rowsLeft[col] / 2;
  }
  puts("   +----------+----------+----------+----------+----------");
  for(row = 0; row < 64; row++)
  {
    printf("%02x:", row);
    for(col = 0; col < 5; col++)
    {
      putchar('|');
      if(rowsLeft[col] >= 0)
      {
        rowsLeft[col]--;
        whichBits = patdata[patbitsptr[col]];
        if(row & 1)
        {
          whichBits >>= 4;
          patbitsptr[col]++;
        }

        if(whichBits & 1)
        {
          if(col == 3) /* noise track omits high nibble of noise */
          {
            gotbyte = 16 - ((patdata[curpatptr[col]] >> nibble[col]) & 0x0f);
            if(nibble[col])
              curpatptr[col]++;
            nibble[col] ^= 4;
          }
          else if(nibble[col])
          {
            gotbyte = (patdata[curpatptr[col]] & 0xf0) |
                     (patdata[curpatptr[col] + 1] & 0x0f);
            curpatptr[col]++;
          }
          else
          {
            gotbyte = patdata[curpatptr[col]];
            curpatptr[col]++;
          }
          if(gotbyte == 97)
            fputs("Off", stdout);
          else if(gotbyte > 97)
            fputs(">97", stdout);
          else if(gotbyte == 0)
            fputs("NUL", stdout);
          else
          {
            gotbyte--;
            fputc(noteNames[gotbyte % 12], stdout);
            fputc(noteSharps[gotbyte % 12], stdout);
            fputc(gotbyte / 12 + '0', stdout);
          }
        }
        else
          fputs("---", stdout);

        if(whichBits & 2)
        {
          gotbyte = ((patdata[curpatptr[col]] >> nibble[col]) & 0x0f) + 1;
          fprintf(stdout, " %02x ", gotbyte);
          if(nibble[col])
            curpatptr[col]++;
          nibble[col] ^= 4;
        }
        else
        {
          fputs(" -- ", stdout);
        }
         
        if(whichBits & 4)
        {
          gotbyte = (patdata[curpatptr[col]] >> nibble[col]) & 0x0f;
          fprintf(stdout, "%X", gotbyte);
          if(nibble[col])
            curpatptr[col]++;
          nibble[col] ^= 4;
          if(whichBits & 8)
          {
            if(nibble[col])
            {
              gotbyte = (patdata[curpatptr[col]] & 0xf0) |
                       (patdata[curpatptr[col] + 1] & 0x0f);
            }
            else
            {
              gotbyte = patdata[curpatptr[col]];
            }
            fprintf(stdout, "%02X", gotbyte);
            curpatptr[col]++;
          }
          else
          {
            fputs("00", stdout);
          }
        }
        else
          fputs("---", stdout);
      }
      else
      {
        fputs("--- -- ---", stdout);
      }
    }
    putchar('\n');
  }
}


int main(int argc, char **argv)
{
  FILE *fp;
  int got;

  if(argc < 2)
  {
    fputs("nt2dump by Damian Yerrick: convert an NT2 NED into text\n"
          "syntax: nt2dump temp.dat\n", stderr);
    return 1;
  }

  fp = fopen(argv[1], "rb");
  if(fp == NULL)
  {
    fputs("nt2dump could not open ", stderr);
    perror(argv[1]);
    return 1;
  }

  if(fread(&hdr, 13, 1, fp) < 1)
  {
    fprintf(stderr, "nt2dump: %s has no header\n", argv[1]);
    fclose(fp);
    return 1;
  }

  if(hdr.version != 0x20 || hdr.headerSize < 13 || hdr.nInsts > 16)
  {
    fprintf(stderr, "nt2dump: %s\nis not a valid temp.dat format NED\n",
            argv[1]);
    fclose(fp);
    return 1;
  }

  PrintHeader();
  if(hdr.headerSize > sizeof(NEDSeqHeader))
  {
    int n = hdr.headerSize - sizeof(NEDSeqHeader);

    do {
      fgetc(fp);
    } while(--n);
  }

  /* Read instruments */
  got = fread(insts, sizeof(NEDInstrument), hdr.nInsts, fp);
  if(got < hdr.nInsts)
  {
    fprintf(stderr, "nt2dump: could only read %d of %d instruments\n", got, (int)hdr.nInsts);
    fclose(fp);
    return 1;
  }
  PrintInsts();

  /* Read orders */
  orders = malloc(hdr.songLength * sizeof(ui32));
  if(orders == NULL)
  {
    fprintf(stderr, "nt2dump: could not allocate RAM for %d orders\n", (int)hdr.songLength);
    fclose(fp);
    return 1;
  }
  got = ReadOrders(orders, hdr.songLength, fp);
  if(got < hdr.songLength)
  {
    fprintf(stderr, "nt2dump: could only read %d of %d order lines\n", got, (int)hdr.songLength);
    free(orders);
    fclose(fp);
    return 1;
  }
  PrintOrders();

  /* Read pattern pointer tables */
  if(ReadPatPtrs(fp))
  {
    fprintf(stderr, "nt2dump: couldn't get all the pattern pointers\n");
    free(orders);
    fclose(fp);
    return 1;
  }

  /* Read pattern data (ONLY WORKS FOR NEDS UP TO 8 KB!) */
  fseek(fp, 0, SEEK_SET);
  fread(patdata, 1, 8192, fp);
  fclose(fp);

  HexDump();

  for(got = 0; got < hdr.songLength; got++)
    PrintTable(got);

  free(orders);
  return 0;
}
